BASIC_FACTS_KEY_MAP = {
    u'os_version': u'kickstart_ver_str',
    u'model': u'chassis_id',
    u'hostname': u'host_name',
    u'serial_number': u'proc_board_id'
}

UPTIME_KEY_MAP = {
    u'up_days': u'kern_uptm_days',
    u'up_hours': u'kern_uptm_hrs',
    u'up_mins': u'kern_uptm_mins',
    u'up_secs': u'kern_uptm_secs'
}

INTERFACE_KEY_MAP = {
    u'description': u'name',
}


VLAN_KEY_MAP = {
    'id': 'vlanshowbr-vlanid-utf',
    'name': 'vlanshowbr-vlanname',
    'state': 'vlanshowbr-vlanstate',
    'admin_state': 'vlanshowbr-shutstate',
}


from pynxos.lib.data_model.converters import converted_list_from_table
from pynxos.lib.data_model.key_maps import VLAN_KEY_MAP
from pynxos.errors import NXOSError

from .base_feature import BaseFeature

#class VlanNotInRangeError(NXOSError):
#    def __init__(self):
#        super(VlanNotInRangeError, self).__init__(
#            'Vlan Id must be in range 1-3967')
#
#def vlan_not_in_range_error(vlan_id):
#    vlan_id = int(vlan_id)
#
#    if vlan_id < 1 or vlan_id > 3967:
#        raise VlanNotInRangeError

class Vlans(BaseFeature):

    def __init__(self, device):
        super(Vlans, self).__init__(device)

#    def get(self, vlan_id):
#        vlan_not_in_range_error(vlan_id)
#
#        vlan_id_table = self.device.show('show vlan id %s' % vlan_id)
#        try:
#            return converted_list_from_table(
#                vlan_id_table, 'vlanbriefid', VLAN_KEY_MAP)[0]
#        except IndexError:
#            return {}

    def get_list(self):
        all_vlan_list = self.get_all()
        vlan_id_list = list(str(x['id']) for x in all_vlan_list)

        return vlan_id_list

    def get_all(self):
        all_vlan_table = self.device.show('show vlan')
        all_vlan_list = converted_list_from_table(
            all_vlan_table, 'vlanbrief', VLAN_KEY_MAP)

        return all_vlan_list

#    def config(self, vlan_id, **params):
#        vlan_not_in_range_error(vlan_id)
#
#        vlan_config_commands = ['vlan %s' % vlan_id]
#        vlan_name = params.get('name')
#        if vlan_name:
#            vlan_config_commands.append('name %s' % vlan_name)
#
#        self.device.config_list(vlan_config_commands)
#
#    def set_name(self, vlan_id, vlan_name=None, default=False, disable=False):
#        vlan_not_in_range_error(vlan_id)
#
#        vlan_config_commands = ['vlan %s' % vlan_id]
#        if vlan_name is None:
#            if default or disable:
#                name_command = 'no name'
#            else:
#                raise NXOSError('vlan_name must be supplied, or default or disable set to True')
#        else:
#            name_command = 'name %s' % vlan_name
#
#        vlan_config_commands.append(name_command)
#        self.device.config_list(vlan_config_commands)
#
#    def remove(self, vlan_id):
#        vlan_not_in_range_error(vlan_id)
#
#        vlan_remove_command = 'no vlan %s' % vlan_id
#        self.device.config(vlan_remove_command)
#
#def instance(device):
#    return Vlans(device)


import sys
import collections

def strip_unicode(data):
    if sys.version_info.major >= 3:
        return data

    if isinstance(data, basestring):
        return str(data)
    elif isinstance(data, collections.Mapping):
        return dict(map(strip_unicode, data.iteritems()))
    elif isinstance(data, collections.Iterable):
        return type(data)(map(strip_unicode, data))
    else:
        return data

def convert_dict_by_key(original, key_map, fill_in=False, whitelist=[], blacklist=[]):
    converted = {}
    for converted_key in key_map:
        original_key = key_map[converted_key]
        if original_key in original:
            converted[converted_key] = original[original_key]
        else:
            converted[converted_key] = None

    if fill_in:
        original_key_subset = []

        if whitelist:
            original_key_subset.extend(list(set(whitelist) - set(key_map.values())))
        else:
            original_key_subset.extend(list(set(original.keys()) - set(blacklist) - set(key_map.values())))

        for original_key in original_key_subset:
            if original_key in original:
                converted[original_key] = original[original_key]

    return converted

def convert_list_by_key(original_list, key_map, fill_in=False, whitelist=[], blacklist=[]):
    converted_list = []
    for original in original_list:
        converted_list.append(
            convert_dict_by_key(original,
                                key_map,
                                fill_in=fill_in,
                                whitelist=whitelist,
                                blacklist=blacklist))

    return converted_list

def list_from_table(table, list_name):
    if table is None:
        return []

    table_key = u'TABLE_%s' % list_name
    row_key = u'ROW_%s' % list_name

    the_list = table[table_key][row_key]

    if not isinstance(the_list, list):
        the_list = [the_list]

    return the_list

def converted_list_from_table(table, list_name, key_map, fill_in=False, whitelist=[], blacklist=[]):
    from_table_list = list_from_table(table, list_name)
    converted_list = convert_list_by_key(from_table_list,
                                         key_map,
                                         fill_in=fill_in,
                                         whitelist=whitelist,
                                         blacklist=blacklist)

    return converted_list


from scp import SCPClient
from pynxos.errors import CLIError, NXOSError

import paramiko
import hashlib
import os
import re

class FileTransferError(NXOSError):
    pass

class FileCopy(object):
    """This class is used to copy local files to a NXOS device.
    """
    def __init__(self, device, src, dst=None, port=22, file_system='bootflash:'):
        self.device = device
        self.src = src
        self.dst = dst or os.path.basename(src)
        self.port = port
        self.file_system = file_system

    def get_flash_size(self):
        """Return the available space in the remote directory.
        """
        dir_out = self.device.show('dir {}'.format(self.file_system), raw_text=True)

        match = re.search(r'(\d+) bytes free', dir_out)
        bytes_free = match.group(1)

        return int(bytes_free)

    def get_remote_size(self):
        return self.get_flash_size()

    def enough_space(self):
        """Check for enough space on the remote device.
        """
        flash_size = self.get_flash_size()
        file_size = os.path.getsize(self.src)
        if file_size > flash_size:
            return False

        return True

    def enough_remote_space(self):
        return self.enough_space()

    def local_file_exists(self):
        return os.path.isfile(self.src)

    def file_already_exists(self):
        """Check to see if there is a remote file with the same
        name and md5 sum.

        Returns:
            ``True`` if exists, ``False`` otherwise.
        """
        dst_hash = self.get_remote_md5()
        src_hash = self.get_local_md5()
        if src_hash == dst_hash:
            return True

        return False

    def already_transfered(self):
        return self.file_already_exists()

    def remote_file_exists(self):
        dir_body = self.device.show(
            'dir {0}/{1}'.format(self.file_system, self.dst), raw_text=True)
        if 'No such file' in dir_body:
            return False

        return True

    def get_remote_md5(self):
        """Return the md5 sum of the remote file,
        if it exists.
        """
        md5_body = self.device.show(
            'show file {0}{1} md5sum'.format(self.file_system, self.dst), raw_text=True)
        return md5_body.strip()

    def get_local_md5(self, blocksize=2**20):
        """Get the md5 sum of the local file,
        if it exists.
        """
        if self.local_file_exists():
            m = hashlib.md5()
            with open(self.src, "rb") as f:
                buf = f.read(blocksize)
                while buf:
                    m.update(buf)
                    buf = f.read(blocksize)
            return m.hexdigest()

    def transfer_file(self, hostname=None, username=None, password=None, pull=False):
        """Transfer the file to the remote device over SCP.

        Note:
            If any arguments are omitted, the corresponding attributes
            of ``self.device`` will be used.

        Args:
            hostname (str): OPTIONAL - The name or
                IP address of the remote device.
            username (str): OPTIONAL - The SSH username
                for the remote device.
            password (str): OPTIONAL - The SSH password
                for the remote device.

        Returns:
            True if successful.

        Raises:
            FileTransferError: if the transfer isn't successful.
        """
        if pull is False:
            if not self.local_file_exists():
                raise FileTransferError(
                    'Could not transfer file. Local file doesn\'t exist.')

            if not self.enough_space():
                raise FileTransferError(
                    'Could not transfer file. Not enough space on device.')

        hostname = hostname or self.device.host
        username = username or self.device.username
        password = password or self.device.password

        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(
            hostname=hostname,
            username=username,
            password=password,
            port=self.port,
            allow_agent=False,
            look_for_keys=False)

        full_remote_path = '{}{}'.format(self.file_system, self.dst)
        scp = SCPClient(ssh.get_transport())
        try:
            if pull:
                scp.get(full_remote_path, self.src)
            else:
                scp.put(self.src, full_remote_path)
        except:
            raise FileTransferError(
                'Could not transfer file. There was an error during transfer. Please make sure remote permissions are set.')
        finally:
            scp.close()

        return True

    def send(self):
        self.transfer_file()

    def get(self):
        self.transfer_file(pull=True)


class NXOSError(Exception):
    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return '%s: %s' % (self.__class__.__name__, self.message)

    __str__ = __repr__

class CLIError(NXOSError):
    def __init__(self, command, message):
        self.command = command
        self.message = message

    def __repr__(self):
        return 'The command "%s" gave the error "%s".' % (self.command, self.message)

    __str__ = __repr__


import signal
import re
from .lib.rpc_client import RPCClient
from .lib import convert_dict_by_key, converted_list_from_table, strip_unicode
from .lib.data_model import key_maps
from pynxos.features.file_copy import FileCopy
from pynxos.features.vlans import Vlans
from pynxos.errors import CLIError, NXOSError


class RebootSignal(NXOSError):
    pass


class Device(object):
    def __init__(self, host, username, password, transport=u'http', port=None, timeout=30, verify=True):
        self.host = host
        self.username = username
        self.password = password
        self.transport = transport
        self.timeout = timeout
        self.verify = verify

        self.rpc = RPCClient(host, username, password, transport=transport, port=port, verify=self.verify)

    def _cli_error_check(self, command_response):
        error = command_response.get(u'error')
        if error:
            command = command_response.get(u'command')
            if u'data' in error:
                raise CLIError(command, error[u'data'][u'msg'])
            else:
                raise CLIError(command, 'Invalid command.')

    def _cli_command(self, commands, method=u'cli'):
        if not isinstance(commands, list):
            commands = [commands]

        rpc_response = self.rpc.send_request(commands, method=method, timeout=self.timeout)

        text_response_list = []
        for command_response in rpc_response:
            self._cli_error_check(command_response)
            text_response_list.append(command_response[u'result'])

        return strip_unicode(text_response_list)

    def show(self, command, raw_text=False):
        """Send a non-configuration command.

        Args:
            command (str): The command to send to the device.

        Keyword Args:
            raw_text (bool): Whether to return raw text or structured data.

        Returns:
            The output of the show command, which could be raw text or structured data.
        """
        commands = [command]
        list_result = self.show_list(commands, raw_text)
        if list_result:
            return list_result[0]
        else:
            return {}

    def show_list(self, commands, raw_text=False):
        """Send a list of non-configuration commands.

        Args:
            commands (list): A list of commands to send to the device.

        Keyword Args:
            raw_text (bool): Whether to return raw text or structured data.

        Returns:
            A list of outputs for each show command
        """
        return_list = []
        if raw_text:
            response_list = self._cli_command(commands, method=u'cli_ascii')
            for response in response_list:
                if response:
                    return_list.append(response[u'msg'])
        else:
            response_list = self._cli_command(commands)
            for response in response_list:
                if response:
                    return_list.append(response[u'body'])

        return return_list

    def config(self, command):
        """Send a configuration command.

        Args:
            command (str): The command to send to the device.

        Raises:
            CLIError: If there is a problem with the supplied command.
        """
        commands = [command]
        list_result = self.config_list(commands)
        return list_result[0]

    def config_list(self, commands):
        """Send a list of configuration commands.

        Args:
            commands (list): A list of commands to send to the device.

        Raises:
            CLIError: If there is a problem with one of the commands in the list.
        """
        return self._cli_command(commands)

    def save(self, filename='startup-config'):
        """Save a device's running configuration.

        Args:
            filename (str): The filename on the remote device.
                If none is supplied, the implementing class should
                save to the "startup configuration".
        """
        try:
            self.show(u'copy run %s' % filename, raw_text=True)
        except CLIError as e:
            if 'overwrite' in e.message:
                return False
            raise

        return True

    def file_copy_remote_exists(self, src, dest=None, file_system='bootflash:'):
        """Check if a remote file exists. A remote file exists if it has the same name
        as supplied dest, and the same md5 hash as the source.

        Args:
            src (str): Path to local file to check.

        Keyword Args:
            dest (str): The destination file path to be saved on remote the remote device.
                If none is supplied, the implementing class should use the basename
                of the source path.
            file_system (str): The file system for the
                remote file. Defaults to 'bootflash:'.

        Returns:
            True if the remote file exists, False if it doesn't.
        """
        fc = FileCopy(self, src, dst=dest, file_system=file_system)
        if fc.file_already_exists():
            return True
        return False

    def file_copy(self, src, dest=None, file_system='bootflash:'):
        """Send a local file to the device.

        Args:
            src (str): Path to the local file to send.

        Keyword Args:
            dest (str): The destination file path to be saved on remote flash.
                If none is supplied, the implementing class should use the basename
                of the source path.
            file_system (str): The file system for the
                remote fle. Defaults to bootflash:'.
        """
        fc = FileCopy(self, src, dst=dest, file_system=file_system)
        fc.send()

    def _disable_confirmation(self):
        self.show('terminal dont-ask')

    def reboot(self, confirm=False):
        """Reboot the device.

        Args:
            confirm(bool): if False, this method has no effect.
        """
        if confirm:
            def handler(signum, frame):
                raise RebootSignal('Interupting after reload')

            signal.signal(signal.SIGALRM, handler)
            signal.alarm(5)

            try:
                self._disable_confirmation()
                self.show('reload')
            except RebootSignal:
                signal.alarm(0)

            signal.alarm(0)
        else:
            print('Need to confirm reboot with confirm=True')

    def set_boot_options(self, image_name, kickstart=None):
        """Set boot variables
        like system image and kickstart image.

        Args:
            The main system image file name.

        Keyword Args: many implementors may choose
            to supply a kickstart parameter to specicify a kickstart image.
        """
        self._disable_confirmation()
        try:
            if kickstart is None:
                self.show('install all nxos %s' % image_name, raw_text=True)
            else:
                self.show('install all system %s kickstart %s' % (image_name, kickstart), raw_text=True)
        except CLIError:
            pass

    def get_boot_options(self):
        """Get current boot variables
        like system image and kickstart image.

        Returns:
            A dictionary, e.g. { 'kick': router_kick.img, 'sys': 'router_sys.img'}
        """
        boot_options_raw_text = self.show('show boot', raw_text=True).split('Boot Variables on next reload')[1]
        if 'kickstart' in boot_options_raw_text:
            kick_regex = r'kickstart variable = bootflash:/(\S+)'
            sys_regex = r'system variable = bootflash:/(\S+)'

            kick = re.search(kick_regex, boot_options_raw_text).group(1)
            sys = re.search(sys_regex, boot_options_raw_text).group(1)
            retdict = dict(kick=kick, sys=sys)
        else:
            nxos_regex = r'NXOS variable = bootflash:/(\S+)'
            nxos = re.search(nxos_regex, boot_options_raw_text).group(1)
            retdict = dict(sys=nxos)

        install_status = self.show('show install all status', raw_text=True)
        retdict['status'] = install_status

        return retdict

    def rollback(self, filename):
        """Rollback to a checkpoint file.

        Args:
            filename (str): The filename of the checkpoint file to load into the running configuration.
        """
        self.show('rollback running-config file %s' % filename, raw_text=True)

    def checkpoint(self, filename):
        """Save a checkpoint of the running configuration to the device.

        Args:
            filename (str): The filename to save the checkpoint as on the remote device.
        """
        self.show_list(['terminal dont-ask', 'checkpoint file %s' % filename], raw_text=True)

    def backup_running_config(self, filename):
        """Save a local copy of the running config.

        Args:
            filename (str): The local file path on which to save the running config.
        """
        with open(filename, 'w') as f:
            f.write(self.running_config)

    @property
    def running_config(self):
        """Return the running configuration of the device.
        """
        response = self.show(u'show running-config', raw_text=True)
        return response

    def _convert_uptime_to_string(self, up_days, up_hours, up_mins, up_secs):
        return '%02d:%02d:%02d:%02d' % (up_days, up_hours, up_mins, up_secs)

    def _convert_uptime_to_seconds(self, up_days, up_hours, up_mins, up_secs):
        seconds = up_days * 24 * 60 * 60
        seconds += up_hours * 60 * 60
        seconds += up_mins * 60
        seconds += up_secs

        return seconds

    def _get_interface_detailed_list(self):
        try:
            interface_table = self.show(u'show interface status')
            interface_list = converted_list_from_table(interface_table, u'interface', key_maps.INTERFACE_KEY_MAP, fill_in=True)
        except CLIError:
            return []

        return interface_list

    def _get_interface_list(self):
        iface_detailed_list = self._get_interface_detailed_list()
        iface_list = list(x['interface'] for x in iface_detailed_list)

        return iface_list

    def _get_vlan_list(self):
        vlans = Vlans(self)
        vlan_list = vlans.get_list()

        return vlan_list

    def _get_show_version_facts(self):
        show_version_result = self.show(u'show version')
        uptime_facts = convert_dict_by_key(show_version_result, key_maps.UPTIME_KEY_MAP)

        up_days = uptime_facts['up_days']
        up_hours = uptime_facts['up_hours']
        up_mins = uptime_facts['up_mins']
        up_secs = uptime_facts['up_secs']

        uptime_string = self._convert_uptime_to_string(up_days, up_hours, up_mins, up_secs)
        uptime_seconds = self._convert_uptime_to_seconds(up_days, up_hours, up_mins, up_secs)

        show_version_facts = convert_dict_by_key(show_version_result, key_maps.BASIC_FACTS_KEY_MAP)

        show_version_facts['uptime'] = uptime_seconds
        show_version_facts['uptime_string'] = uptime_string

        return show_version_facts

    @property
    def facts(self):
        """Return a dictionary of facts about the device.

        The dictionary must include the following keys.
        All keys are strings, the value type is given in parenthesis:
            uptime (int)
            vendor (str)
            os_version (str)
            interfaces (list of strings)
            hostname (str)
            fqdn (str)
            uptime_string (str)
            serial_number (str)
            model (str)
            vlans (list of strings)

        The dictionary can also include a vendor-specific dictionary, with the
        device type as a key in the outer dictionary.

        Example:
            {
                "uptime": 1819711,
                "vendor": "cisco",
                "os_version": "7.0(3)I2(1)",
                "interfaces": [
                    "mgmt0",
                    "Ethernet1/1",
                    "Ethernet1/2",
                    "Ethernet1/3",
                    "Ethernet1/4",
                    "Ethernet1/5",
                    "Ethernet1/6",
                ],
                "hostname": "n9k1",
                "fqdn": "N/A",
                "uptime_string": "21:01:28:31",
                "serial_number": "SAL1819S6LU",
                "model": "Nexus9000 C9396PX Chassis",
                "vlans": [
                    "1",
                    "2",
                    "3",
                ]
            }
        """
        if hasattr(self, '_facts'):
            return self._facts

        facts = {}

        show_version_facts = self._get_show_version_facts()
        facts.update(show_version_facts)

        iface_list = self._get_interface_list()
        facts['interfaces'] = iface_list

        vlan_list = self._get_vlan_list()
        facts['vlans'] = vlan_list

        facts['fqdn'] = 'N/A'

        self._facts = facts
        return facts

#    def feature(self, feature_name):
#        try:
#            feature_module = importlib.import_module(
#                'pynxos.features.%s' % (feature_name))
#            return feature_module.instance(self)
#        except ImportError:
#            raise FeatureNotFoundError(feature_name, self.device_type)



import requests
from requests.auth import HTTPBasicAuth
import json

from builtins import range
from pynxos.errors import NXOSError

# requests.packages.urllib3.disable_warnings()

class RPCClient(object):
    def __init__(self, host, username, password, transport=u'http', port=None, verify=True):
        if transport not in ['http', 'https']:
            raise NXOSError('\'%s\' is an invalid transport.' % transport)

        if port is None:
            if transport == 'http':
                port = 80
            elif transport == 'https':
                port = 443

        self.url = u'%s://%s:%s/ins' % (transport, host, port)
        self.headers = {u'content-type': u'application/json-rpc'}
        self.username = username
        self.password = password
        self.verify = verify

    def _build_payload(self, commands, method, rpc_version=u'2.0'):
        payload_list = []

        id_num = 1
        for command in commands:
            payload = dict(jsonrpc=rpc_version,
                           method=method,
                           params=dict(cmd=command, version=1),
                           id=id_num,)

            payload_list.append(payload)
            id_num += 1

        return payload_list

    def send_request(self, commands, method=u'cli', timeout=30):
        timeout=int(timeout)
        payload_list = self._build_payload(commands, method)
        response = requests.post(self.url,
                                 timeout=timeout,
                                 data=json.dumps(payload_list),
                                 headers=self.headers,
                                 auth=HTTPBasicAuth(self.username, self.password),
                                 verify=self.verify)

        response_list = json.loads(response.text)

        if isinstance(response_list, dict):
            response_list = [response_list]

        for i in range(len(commands)):
            response_list[i][u'command'] = commands[i]

        return response_list



class BaseFeature(object):

    def __init__(self, device):
        self.device = device

    def get(self, vlan_id):
        raise NotImplementedError

    def get_list(self):
        raise NotImplementedError

    def get_all(self):
        raise NotImplementedError

    def config(self, vlan_id, **params):
        raise NotImplementedError

    def remove(self, vlan_id):
        raise NotImplementedError